Let's run a NodeJS express server at port 3002 on the host computer so that visiting localhost:3002/ will show a json message "message": "Hello world"

What's great about Docker is that you can build the image, then run a virtual container off that image at another port than the usual port 3000, and that's very convenient for developers you distribute the code to or to other servers you migrate to. Port 3000, 3001, 3002, etc may be taken and you only have port 3003 available. By running a port number at the `docker run` command when it's time to spin up the server, it's quite flexible to the port situation.

Again, let's develop it locally first

---

.dockerignore:
```
.DS_Store
node_modules
package-lock.json
__pycache__
*.pyc
.env
```

Dockerfile:
```
# Use the specific NodeJS base image (bundled in Debian)
FROM node:18

# Set the working directory in the container.
# All below commands will work from working directory.
WORKDIR /app

# Copy the package.json file to the container
COPY package.json .

# Install the NodeJS packages
RUN npm install

# Copy your application files (if any)
COPY . .

# Specify the command to run your application (optional, depends on your use case)
CMD ["node", "app.js"]
```

package.json:
```
{
  "dependencies": {
    "express": "4.21.2"
  }
}
```

app.js:
```
const express = require('express');
const app = express();

app.get('/', (req, res) => {
  res.json({ message: "Hello world" });
});

const port = 3000;
app.listen(port, () => {
  console.log(`Server running on port ${port}`);
});
```

---

Let's build the image:
```
docker build -t express-app .
```

Then we run it on port 3002 or your choice of port:
```
docker run -p 3002:3000 -d express-app
```

Main points of running port:
- The port positions are `host:container` . The container port must be 3000 because that's the port that `app.js` is coded to listen at.
- -d is daemon mode (aka background mode to layperson, aka detached mode to Docker fans) so doesn't hog up your terminal (nor does leaving the terminal close the session out). Common use case because you may want to spin the server then leave it running 24/7 as you log out the SSH terminal session.

Test the server:
Visit `http://localhost:3002` on your computer to check it works. You should see a "Hello world" message in JSON format:

![[Pasted image 20250303004849.png]]

Introspection:
If you hadn't ran with `-d` flag, the server would be tied to your terminal session. Since it IS in daemon mode (aka background mode to layperson, aka detached mode to Docker fans), you may easily forget the container is still running.

The same rules still apply about listing what containers are running, regardless if they're hogging the terminal are in daemon mode. Run `docker ps` to see all active containers.

Your output for express-app should look similar to:
```
CONTAINER ID   IMAGE                                   COMMAND                  CREATED         STATUS         PORTS                    NAMES  
377f9be6ece4   express-app                             "docker-entrypoint.s…"   6 minutes ago   Up 6 minutes   0.0.0.0:3002->3000/tcp   infallible_black  
```

Cleanup locally:
You can close off a container by the container id like so:
```
docker kill 377f9be6ece4
```

^ Note you do not kill off based on the image name because the container is a running virtual instance. The idea is that an image can be spun up to multiple containers as necessary (for example, splitting load between multiple port numbers as traffic increases). Tools like **Docker Compose** or **Kubernetes** can automate scaling and orchestration, which we may cover in a future challenge series.

Test remotely:
Copy your folder of files into a remote server. Then SSH and repeat the steps to check it's the same output when you build then run the image. 

If you have multiple 300X ports running, you may want to listen for which ports are running, 
1. Find out what distro you're server is running on: `cat /etc/os-release`.
2. Eg on Ubuntu 22, you can test certain ports (succeeded means it's used, refused means it's available): `nc -zv localhost PORT_NUMBER`. Find the port that's available
3. As a quick review, you may be running the commands:
```
docker build -t express-app .
docker run -p AVAILABLE_PORT:3000 -d express-app
```

4. Then visit your domain at `http://YOURDOMAIN.com:3020/`
You could get:
![[Pasted image 20250303011251.png]]

Issues? If it didn't work, confirm that the container is running with `docker ps`. If behind a firewall, you may have to allow the port either through command line or your web hotsing panel if applicable. Refer to [[UFW - Enable specific ports]] if that's the firebase you're using.

Note that in production, you don't want your application URL to be at http://DOMAIN.com:PORT because it looks unprofessional, is not SEO-friendly, and is telling hackers what ports are exposed. In production, you would block that port behind a firewall but at apache or nginx level, perform a reverse lookup from a friendlier url (Like https://DOMAIN.com/app) to the port internally. This would allow SSL/https certification to apply at the app as well.

You may want to perform the cleanup steps at the remote server as well.